/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include "../../../include/gstaudio.h"
#include "volume.h"



static GstElementDetails volume_details = {
  "Volume",
  "Filter/Effect",
  "Set volume on mono or stereo audio/raw streams",
  VERSION,
  "Andy Wingo <apwingo@eos.ncsu.edu>",
  "(C) 2001",
};


/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_MUTED,
  ARG_VOLUME,
  ARG_LEFT,
  ARG_RIGHT
};

static GstPadTemplate*
volume_src_factory (void)
{
  static GstPadTemplate *template = NULL;

  if (!template) {
    template = gst_padtemplate_new (
      "src",
      GST_PAD_SRC,
      GST_PAD_ALWAYS,
      gst_caps_new (
        "volume_src",
        "audio/raw",
	GST_AUDIO_INT_PAD_TEMPLATE_PROPS,
	NULL);
  }
  return template;
}

static GstPadTemplate*
volume_sink_factory (void)
{
  static GstPadTemplate *template = NULL;

  if (!template) {
    template = gst_padtemplate_new (
      "sink",
      GST_PAD_SINK,
      GST_PAD_ALWAYS,
      gst_caps_new (
        "test_src",
        "audio/raw",
	GST_AUDIO_INT_PAD_TEMPLATE_PROPS,
      NULL);
  }
  return template;
}

static void		gst_volume_class_init		(GstVolumeClass *klass);
static void		gst_volume_init			(GstVolume *filter);

static void		gst_volume_set_property			(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void		gst_volume_get_property			(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

static void		gst_volume_chain			(GstPad *pad, GstBuffer *buf);
static void inline 	gst_volume_fast_16bit_chain_stereo 	(gint16* data, gint16** out_data, 
			         				 guint numsamples, GstVolume *filter);
static void inline 	gst_volume_fast_8bit_chain_stereo	(gint8* data, gint8** out_data,
			         				 guint numsamples, GstVolume *filter);
static void inline 	gst_volume_fast_16bit_chain_mono 	(gint16* data, gint16** out_data, 
			         				 guint numsamples, GstVolume *filter);
static void inline 	gst_volume_fast_8bit_chain_mono	        (gint8* data, gint8** out_data,
			         				 guint numsamples, GstVolume *filter);

static GstElementClass *parent_class = NULL;
//static guint gst_filter_signals[LAST_SIGNAL] = { 0 };

static GstPadNegotiateReturn
volume_negotiate_src (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstVolume* filter = GST_VOLUME (gst_pad_get_parent (pad));
  
  if (*caps==NULL) 
    return GST_PAD_NEGOTIATE_FAIL;
  
  return gst_pad_negotiate_proxy(pad,filter->sinkpad,caps);
}

static GstPadNegotiateReturn
volume_negotiate_sink (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstVolume* filter = GST_VOLUME (gst_pad_get_parent (pad));
  
  if (*caps==NULL) 
    return GST_PAD_NEGOTIATE_FAIL;
  
  return gst_pad_negotiate_proxy(pad,filter->srcpad,caps);
}		

GType
gst_volume_get_type(void) {
  static GType volume_type = 0;

  if (!volume_type) {
    static const GTypeInfo volume_info = {
      sizeof(GstVolumeClass),      NULL,
      NULL,
      (GClassInitFunc)gst_volume_class_init,
      NULL,
      NULL,
      sizeof(GstVolume),
      0,
      (GInstanceInitFunc)gst_volume_init,
    };
    volume_type = g_type_register_static(GST_TYPE_ELEMENT, "GstVolume", &volume_info, 0);
  }
  return volume_type;
}

static void
gst_volume_class_init (GstVolumeClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_SILENT,
    g_param_spec_boolean("silent","silent","silent",
                         TRUE,G_PARAM_READWRITE));

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_MUTED,
    g_param_spec_boolean("muted","muted","muted",
                         FALSE,G_PARAM_READWRITE));

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_VOLUME,
    g_param_spec_float("volume","volume","volume",
                       -4.0,4.0,1.0,G_PARAM_READWRITE));

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_LEFT,
    g_param_spec_float("left","left","left",
                       -4.0,4.0,1.0,G_PARAM_READWRITE));

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_RIGHT,
    g_param_spec_float("right","right","right",
                       -4.0,4.0,1.0,G_PARAM_READWRITE));

  gobject_class->set_property = gst_volume_set_property;
  gobject_class->get_property = gst_volume_get_property;
}

static void
gst_volume_init (GstVolume *filter)
{
  filter->sinkpad = gst_pad_new_from_template(volume_sink_factory (),"sink");
  filter->srcpad = gst_pad_new_from_template(volume_src_factory (),"src");

  gst_pad_set_negotiate_function(filter->sinkpad,volume_negotiate_sink);
  gst_pad_set_negotiate_function(filter->srcpad,volume_negotiate_src);

  gst_element_add_pad(GST_ELEMENT(filter),filter->sinkpad);
  gst_pad_set_chain_function(filter->sinkpad,gst_volume_chain);
  filter->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(filter),filter->srcpad);

  filter->caps_given = FALSE;
  filter->silent = TRUE;
  filter->muted = FALSE;
  filter->volume_i = 8192;
  filter->volume_f = 1.0;
  filter->left_i  = 8192;
  filter->right_i = 8192;
  filter->left_f  = 1.0;
  filter->right_f = 1.0;
}

static void
gst_volume_chain (GstPad *pad, GstBuffer *buf)
{
  GstVolume *filter;
  gint16 *in_data;
  gint16 *out_data;
  GstBuffer* outbuf;
  guint width, rate, channels;
  guint endianness;
  gboolean is_signed;

  GstCaps *caps;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  filter = GST_VOLUME(GST_OBJECT_PARENT (pad));
  g_return_if_fail(filter != NULL);
  g_return_if_fail(GST_IS_VOLUME(filter));

  caps = NULL;
  caps = GST_PAD_CAPS(pad);
  
  if (caps == NULL)
  {
      g_warning ("volume chain : Could not get caps of pad !");
  }

  /* print out caps of sink pad */

  width      = gst_caps_get_int (caps, "width");
  rate       = gst_caps_get_int (caps, "rate");
  channels   = gst_caps_get_int (caps, "channels");
  endianness = gst_caps_get_int (caps, "endianness");
  is_signed  = gst_caps_get_int (caps, "signed");
  if (!filter->caps_given)
  {
    if (!filter->silent)
    {
      g_print ("Volume : bit width %d, channels %d, rate %d\n",  
      		width, channels, rate);
      g_print ("Volume : endianness %d, signed %s\n",
    		endianness, is_signed ? "yes" : "no");
    }
    filter->caps_given = TRUE;
  }

  in_data = (gint16 *)GST_BUFFER_DATA(buf);
  outbuf=gst_buffer_new();
  GST_BUFFER_DATA(outbuf) = (gchar*)g_new(gint16,GST_BUFFER_SIZE(buf)/2);
  GST_BUFFER_SIZE(outbuf) = GST_BUFFER_SIZE(buf);

  out_data = (gint16*) GST_BUFFER_DATA(outbuf);
  
  if (channels==2) {
      switch (width) {
      case 16:
          gst_volume_fast_16bit_chain_stereo(in_data,&out_data,GST_BUFFER_SIZE(buf)/2, filter);
          break;
      case 8:
          gst_volume_fast_8bit_chain_stereo((gint8*)in_data,(gint8**)&out_data,GST_BUFFER_SIZE(buf), filter);
          break;
      }
  } else {
      switch (width) {
      case 16:
          gst_volume_fast_16bit_chain_mono(in_data,&out_data,GST_BUFFER_SIZE(buf)/2, filter);
          break;
      case 8:
          gst_volume_fast_8bit_chain_mono((gint8*)in_data,(gint8**)&out_data,GST_BUFFER_SIZE(buf), filter);
          break;
      }
  }
  
  gst_buffer_unref(buf);
  gst_pad_push(filter->srcpad,outbuf);
}

static void inline
gst_volume_fast_16bit_chain_stereo(gint16* in_data, gint16** out_data, 
                                   guint numsamples, GstVolume *filter)
#include "stereo-filter.func"

static void inline
gst_volume_fast_8bit_chain_stereo(gint8* in_data, gint8** out_data,
                                   guint numsamples, GstVolume *filter)
#include "stereo-filter.func"

static void inline
gst_volume_fast_16bit_chain_mono(gint16* in_data, gint16** out_data, 
                                   guint numsamples, GstVolume *filter)
#include "mono-filter.func"

static void inline
gst_volume_fast_8bit_chain_mono(gint8* in_data, gint8** out_data,
                                guint numsamples, GstVolume *filter)
#include "mono-filter.func"

static void
gst_volume_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GstVolume *filter;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VOLUME(object));
  filter = GST_VOLUME(object);

  switch (prop_id) 
  {
  case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
  case ARG_MUTED:
      filter->muted = g_value_get_boolean (value);
      break;
  case ARG_VOLUME:
      filter->volume_f       = g_value_get_float (value);
      filter->left_f  = filter->volume_f;
      filter->right_f = filter->volume_f;
      filter->volume_i       = (gint16) filter->volume_f*8192;
      filter->left_i  = (gint16) filter->left_f*8192;
      filter->right_i = (gint16) filter->right_f*8192;
      break;
  case ARG_LEFT:
      filter->left_f  = g_value_get_float (value);
      filter->left_i  = (gint16) filter->left_f*8192;
      filter->volume_f       = (ABS(filter->left_f) +
                                ABS(filter->right_f))/2;
      filter->volume_i       = (gint16) filter->volume_f*8192;
      break;
  case ARG_RIGHT:
      filter->right_f = g_value_get_float (value);
      filter->right_i = (gint16) filter->right_f*8192;
      filter->volume_f       = (ABS(filter->left_f) +
                                ABS(filter->right_f))/2;
      filter->volume_i       = (gint16) filter->volume_f*8192;
      break;
  default:
      break;
  }
}

static void
gst_volume_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GstVolume *filter;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VOLUME(object));
  filter = GST_VOLUME(object);

  switch (prop_id) {
  case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
  case ARG_MUTED:
      g_value_set_boolean (value, filter->muted);
      break;
  case ARG_VOLUME:
      g_value_set_float (value, filter->volume_f);
      break;
  case ARG_LEFT:
      g_value_set_float (value, filter->left_f);
      break;
  case ARG_RIGHT:
      g_value_set_float (value, filter->right_f);
      break;
  default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_elementfactory_new("volume",GST_TYPE_VOLUME,
                                   &volume_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  
  gst_elementfactory_add_padtemplate (factory, volume_src_factory ());
  gst_elementfactory_add_padtemplate (factory, volume_sink_factory ());

  gst_plugin_add_factory(plugin,factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "volume",
  plugin_init
};
